#ifndef _APD_H
#define _APD_H

#include <glib.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <errno.h>
#include <stdarg.h>
#include <stdbool.h>
#ifndef _HAVE_UINT
#   define _HAVE_UINT
    typedef unsigned int uint;
#endif

// ----- Defaults -----
#define APD_DEFAULT_REPORT_THRESHOLD  2
#define APD_DEFAULT_ASSERT_THRESHOLD  5
#define APD_DEFAULT_MAX_FUNCTION_DEPTH  200
#define APD_DEFAULT_ERROR_STREAM  stdout
#define APD_DEFAULT_REPORT_THRESHOLD_DECREASES_WITH_STACK  0 /*false*/

// ----- Function description struct -----
struct apd_function_description_struct {
  int importance;  /* only used when reporting function stacks */
  char *name;
  char *description;
  };

// ---- Main debug data structure ----
struct apd_object_struct {
    bool internal_debug;            // Active internal_debug routines if 1 (and if they are compiled in)
    int internal_debug_level;       // Print APD_INTERNAL_REPORTs if they have an importance  <= this
    bool check_internal;            // Constantly check consistensy of internal data structures 
                                    //     if 1 (and they are compiled in)
    bool action_extra_info;         // add file, line, function info to actions
    bool report_threshold_decreases_with_stack;  
                                    // Decrease the report threshold each time the stack increases if 1
    int report_threshold;
    int assert_threshold;
    bool valgrind_stack_trace;      // Add a valgrind stack trace to apd stack traces if this is 1
    uint max_function_depth;
    GSList * description_stack;
    uint description_stack_size;
    char *current_function_name;    // Points to the function name in the current top function description struct
    FILE *error_stream;
    struct apd_object_struct *backup;  // mirror copy of main struct to detect memory corruption
    char **level_spacers;           // A string to prepend to the report strings for each level
};

// APD_DEBUG_COMPILE is a shortcut that compiles in assertion checking and debug reporting
#ifdef APD_DEBUG_COMPILE
#    define APD_ASSERT_COMPILE
#    define APD_REPORT_COMPILE
#endif
// _APD_EITHER_COMPILE should be on if either APD_ASSERT_COMPILE or APD_REPORT_COMPILE
// are defined.
#if defined(APD_ASSERT_COMPILE)
#    define _APD_EITHER_COMPILE
#elif defined(APD_REPORT_COMPILE)
#    define _APD_EITHER_COMPILE
#endif

// Valgrind will give a nice stack trace when it catches a segfault.
// If this setting is 1, cause a segfault at the end of APD_DIE or after
// assertion failures.
#define APD_SEGFAULT_ON_DIE  1

// Leave a trace in the code of which compile options were enabled
// These variables are just markers, they are never used.
#ifdef APD_ASSERT_COMPILE
     extern int _apd_assert_compile_defined; /* Debug marker only */
#else
     extern int _apd_assert_compile_not_defined; /* Debug marker only */
#endif

#ifdef APD_REPORT_COMPILE
     extern int _apd_report_compile_defined; /* Debug marker only */
#else
     extern int _apd_report_compile_not_defined; /* Debug marker only */
#endif

// When not compiled in, APD_ASSERT_ALWAYS, APD_REPORT_ALWAYS, and family are
// still active.  They will use this error stream when not compiled in.
#ifndef APD_DEFAULT_ERROR_STREAM
#    define APD_DEFAULT_ERROR_STREAM  stdout
#endif

// APD_ADJUST_REPORT_THRESHOLD adjusts the relative importance of debug report messages
// for a particular file (or section of code if it is redefined in the middle of code).
// Set to 0 for no adjustment.  Set to 1 to adjust the report threshold one
// higher than before (you'll see more messages).
// APD_ADJUST_ASSERT_THRESHOLD is the same thing for assertions.
// If you want to avoid compiler warnings, use #undef before #define when
// redefining these after including apd.h.
#ifndef APD_ADJUST_REPORT_THRESHOLD
#    define APD_ADJUST_REPORT_THRESHOLD 0
#endif
#ifndef APD_ADJUST_ASSERT_THRESHOLD
#    define APD_ADJUST_ASSERT_THRESHOLD 0
#endif

// APD_IMPORTANCE_ABSOLUTE_LIMIT defines the absolute limit of importance.  This is
// useful for assertions inside the debug library that make sure importance is within
// some reasonable limits.  Just make sure your importance range is less than this.
#ifndef APD_IMPORTANCE_ABSOLUTE_LIMIT
#    define APD_IMPORTANCE_ABSOLUTE_LIMIT  20
#endif

// APD_INTERNAL_DEBUG_COMPILE compiles in the possibility of three kinds of debugging for
// the internal_debug library itself.  Each of the two kinds of debugging (reports and assertions)
// must be turned
// on at runtime, but just having them compiled in will slow down execution slightly.
// Leave this on unless you have a reason to turn it off.
// Set APD_INTERNAL_DEBUG_COMPILE to 0 to not compile internal debugging.
#ifndef APD_INTERNAL_DEBUG_COMPILE
#    define APD_INTERNAL_DEBUG_COMPILE 1
#endif

// Function names are prepended by characters to indicate the depth
// of the function call.  This is the character used:
#ifndef APD_LEVEL_SPACER_CHARACTER
#    define APD_LEVEL_SPACER_CHARACTER '|'
#endif

// There must be a central place to hold data on which functions have been called,
// what their debug level is, and so on.  This can be accomplished two ways:
// 1) Leave it up to the user to define a variable to hold this information.
// This is more flexible, but requires that you define the debug object and make
// sure it is always in scope.
// 2) Use a global variable in the apd library.
// This is easy (and the default), and the best way if you are not using threads or have some other
// reason to control the debug object.
// Define APD_USE_INTERNAL_DEBUG_OBJECT to 0 if you want to control
// the debug object in your code:
#ifndef APD_USE_INTERNAL_DEBUG_OBJECT
#    define APD_USE_INTERNAL_DEBUG_OBJECT 1
#endif
#if APD_USE_INTERNAL_DEBUG_OBJECT == 1
    // This global is instantiated in apd.c
    extern struct apd_object_struct *apd_object;
#endif

// A string representing the current function must be provided by the compiler.  gcc has
// done this for a while using the __FUNCTION__ variable, and the C99 standard uses
// __func__ to do the same thing.  Pick which one to use here:
// ISO C99 standard:
#if __STDC_VERSION__ >= 199901L
#    define APD_FUNCTION_NAME __func__
// GNU C compilers:
#elif __GNUC__
#    define APD_FUNCTION_NAME  __func__
#else
#    define APD_FUNCTION_NAME "unknown function"
#endif

// If you are using a compiler that doesn't support the C99 "__func__" variable, you
// can also add code like this before each function:
// #undef APD_FUNCTION_NAME
// #define APD_FUNCTION_NAME "my_function"

// Define this to any code you want to be performed regularly (actually,
// whenever macros from this library are called).  This code is run
// in report and assertion statements if they are compiled in.  If not compiled
// in, this code will not be run.
#ifndef APD_CHECK_REGULARLY
#    define APD_CHECK_REGULARLY /* don't check anything regularly */
#endif

// Some things have to be changed to make test mode work.  For example, we want
// to be able to test the assertion failures without aborting the test.  Define APD_TEST_MODE
// to change aborts to printing "<abort>"
#ifdef _APD_TEST_MODE
#   define _APD_ABORT(error_stream, exit_code) \
          fprintf(APD_DEFAULT_ERROR_STREAM, "?");  fflush(APD_DEFAULT_ERROR_STREAM)

#elif APD_SEGFAULT_ON_DIE == 1
#   define _APD_ABORT(error_stream, exit_code) \
        fprintf(error_stream, "\nIntentionally causing segfault for debugger to catch\n"); \
        fflush(error_stream); \
        *((uint *)(0)) = 0; \
        fprintf(error_stream, "\nSegfault failed, exiting.\n"); \
        exit(exit_code)

#elif APD_SEGFAULT_ON_DIE == 0
#   define _APD_ABORT(error_stream, exit_code) \
        exit(exit_code)
#endif

// When assertions or reports are on, APD_PUSH_FUNCTION() is used
// to add a function description, name, and importance to the function info
// stack.  APD_REPORT_FUNCTION_DESCRIPTION() can be used at any place in
// the program to report the current function stack.
//
// Errata: the following was changed because gcc >3.0 doesn't allow it any more:
// The "" is added before APD_FUNCTION_NAME in case APD_FUNCTION_NAME is undefined.
// The library function handles the error if that happens.
// These routines should be on if either assertions or reports are on.
#ifdef _APD_EITHER_COMPILE
  #define APD_PUSH_FUNCTION(importance, ...) \
      APD_CHECK_REGULARLY \
      _apd_push_call_description(apd_object, importance - APD_ADJUST_REPORT_THRESHOLD, \
            __FILE__, __LINE__, APD_FUNCTION_NAME, "" __VA_ARGS__)
  #define APD_POP_FUNCTION() \
      APD_CHECK_REGULARLY \
      _apd_pop_call_description(apd_object, __FILE__, __LINE__, APD_FUNCTION_NAME)
  #define APD_REPORT_FUNCTION_STACK(threshold) \
      APD_CHECK_REGULARLY \
     _apd_report_function_stack(apd_object, threshold + APD_ADJUST_REPORT_THRESHOLD)
  #define APD_REPORT_FUNCTION_STACK_ALL() \
      APD_CHECK_REGULARLY \
     _apd_report_function_stack(apd_object, APD_IMPORTANCE_ABSOLUTE_LIMIT)
#else
  #define APD_PUSH_FUNCTION(...) /* function tracking off */
  #define APD_POP_FUNCTION(...)  /* function tracking off */
  #define APD_REPORT_FUNCTION_STACK(...)    /* function tracking off */
  #define APD_REPORT_FUNCTION_STACK_ALL()  /* function tracking off */
#endif

// _APD_IF_THRESHOLD_EXCEEDED and _APD_REPORT_GUTS() are internal macros used by other macros below
// Note that _APD_IF_THRESHOLD_EXCEEDED can be used even if there is no apd_object, but only if the
// importance is 0.  That is why there is an extra check for 0 at the beginning
#define _APD_IF_THRESHOLD_EXCEEDED(importance, comparison, threshold) \
    if ((importance == 0) || ((importance - APD_ADJUST_REPORT_THRESHOLD) comparison threshold))

// Several places in this file, a debug string will need to give the file, function,
// and line number.  This is the string passed to fprintf() to report this information.
#define APD_FUNCTION_NAME_FORMAT  "%-55s"
#define APD_BEFORE_FILE_LINE_FORMAT " "
#define APD_FILE_LINE_FORMAT  "<%s:%4d>"

// This is the internal function that generates a report line for all the other
// functions except _APD_DIE().  Change this if you want to change the
// way the reports look.
#define _APD_PRINT_APD_REPORT(error_stream, function, file, line, ...) \
      fprintf(error_stream, APD_FUNCTION_NAME_FORMAT, function); \
      fprintf(error_stream, "" __VA_ARGS__); \
      fprintf(error_stream, APD_BEFORE_FILE_LINE_FORMAT  APD_FILE_LINE_FORMAT "\n", file, line); \
      fflush(error_stream)

#define _APD_REPORT_GUTS(importance, comparison, threshold, ...) \
      do { \
          _APD_IF_THRESHOLD_EXCEEDED(importance, comparison, threshold) \
            { \
              _APD_PRINT_APD_REPORT(apd_object->error_stream, apd_object->current_function_name, __FILE__, __LINE__, __VA_ARGS__); \
            } \
         } while (0) /* swallow semicolon */

// APD_REPORT sends all the __VA_ARGS__ to vfprintf for printing if the
// report threshold is greater or equal to the reporting importance
#ifdef APD_REPORT_COMPILE
#    define APD_REPORT(importance, ...) \
        APD_CHECK_REGULARLY \
        _APD_REPORT_GUTS(importance, <=, apd_object->report_threshold, __VA_ARGS__)
#else
#    define APD_REPORT(importance, ...)
#endif

// APD_REPORT_OTHERWISE reports when APD_REPORT does
// not (if the report threshold is less than the reporting importance)
#ifdef APD_REPORT_COMPILE
#    define APD_REPORT_OTHERWISE(importance, ...)  \
          APD_CHECK_REGULARLY \
          _APD_REPORT_GUTS(importance, >, apd_object->report_threshold, __VA_ARGS__)
#else
#    define APD_REPORT_OTHERWISE(importance, ...)
#endif

// APD_REPORT_ALWAYS reports all the time.  If reporting is not compiled in, it just uses fprintf
// to default error stream
#ifdef APD_REPORT_COMPILE
#    define APD_REPORT_ALWAYS(...)  \
          APD_CHECK_REGULARLY \
          APD_REPORT(0, __VA_ARGS__)
#else
#    define APD_REPORT_ALWAYS(...) \
          _APD_PRINT_APD_REPORT(APD_DEFAULT_ERROR_STREAM, APD_FUNCTION_NAME, __FILE__, __LINE__, "" __VA_ARGS__)
#endif

// APD_REPORT_ACTION executes all the statements after the importance parameter
// if the threshold is greater or equal to the reporting importance
#ifdef APD_REPORT_COMPILE
#    define APD_REPORT_ACTION(importance, ...) \
        do \
          { \
            APD_CHECK_REGULARLY \
            _APD_IF_THRESHOLD_EXCEEDED(importance, <=, apd_object->report_threshold) { \
                if (apd_object->action_extra_info) { \
                    _APD_PRINT_APD_REPORT(APD_DEFAULT_ERROR_STREAM, APD_FUNCTION_NAME, \
                                          __FILE__, __LINE__, "Report action here."); } \
                __VA_ARGS__ } \
          } \
        while (0) /*swallow semicolon*/
#else
#    define APD_REPORT_ACTION(importance, ...)
#endif

// APD_*_ACTION_ALWAYS wouldn't make sense.  That would just be code.

// APD_ASSERT tests a condition, and fails with a function description trace.
// APD_ASSERT does nothing when debug mode is off.
// NOTE: the variables arguments at the end must start with a string constant
// to avoid a compiler syntax error.
#define APD_ASSERT_GUTS(importance, test_expression, error_stream, execute_me, ...) \
    do \
      { \
        _APD_IF_THRESHOLD_EXCEEDED(importance, <=, apd_object->assert_threshold) \
           { \
             if ( !(test_expression) ) \
               { \
                  _APD_PRINT_APD_REPORT(error_stream, APD_FUNCTION_NAME, __FILE__, __LINE__, \
                        "Assertion failed: " #test_expression "\n" __VA_ARGS__); \
                  execute_me \
                  _APD_ABORT(error_stream, EXIT_FAILURE); \
               } \
           } \
      } while (0) /* swallow semicolon */

#ifdef APD_ASSERT_COMPILE
#    define APD_ASSERT(importance, test_expression, ...) \
          APD_CHECK_REGULARLY \
          APD_ASSERT_GUTS(importance, test_expression, apd_object->error_stream, \
                APD_REPORT_FUNCTION_STACK_ALL(); , __VA_ARGS__);
#else
#    define APD_ASSERT(importance, test_expression, ...)
#endif

// APD_ACTION executes C code if the threshold is greater or equal to the assertion importance.
#ifdef APD_ASSERT_COMPILE
#    define APD_ACTION(importance, ...) \
        do \
          { \
            APD_CHECK_REGULARLY \
            _APD_IF_THRESHOLD_EXCEEDED(importance, <=, apd_object->assert_threshold) { \
                if (apd_object->action_extra_info) { \
                    _APD_PRINT_APD_REPORT(APD_DEFAULT_ERROR_STREAM, APD_FUNCTION_NAME, __FILE__, __LINE__, "Assert action here."); } \
                __VA_ARGS__ } \
          } \
        while (0) /*swallow semicolon*/
#else
#    define APD_ACTION(importance, ...)
#endif

// APD_ASSERT_ACTION tests the test_expression assertion if the threshold is greater or equal to the assertion importance.
// The actions are performed if the assertion fails.
#ifdef APD_ASSERT_COMPILE
#    define APD_ASSERT_ACTION(importance, test_expression, execute_me) \
        do \
          { \
            APD_CHECK_REGULARLY \
            APD_ASSERT_GUTS(importance, test_expression, apd_object->error_stream, \
                execute_me \
                APD_REPORT_FUNCTION_STACK_ALL(); , ""); \
          } \
        while (0) /*swallow semicolon*/
#else
#    define APD_ASSERT_ACTION(importance, ...)
#endif

// APD_ASSERT_ALWAYS is used for assertions that must always be tested, whether
// or not we are in debug mode, and even if debugging is not compiled in.
// If debugging is not compiled in and on, APD_ASSERT_ALWAYS doesn't
// give a function description trace (for obvious reasons).
#ifdef APD_ASSERT_COMPILE
#    define APD_ASSERT_ALWAYS(test_expression, ...) \
          APD_CHECK_REGULARLY \
          APD_ASSERT(0, test_expression, __VA_ARGS__)
#else
#    define APD_ASSERT_ALWAYS(test_expression, ...) \
          APD_ASSERT_GUTS(0, test_expression, APD_DEFAULT_ERROR_STREAM, /* do nothing */, __VA_ARGS__)
#endif

// APD_ASSERT_ALWAYS_LIBRARY is used for assertions within library packages.  
// The difference between it and the normall APD_ASSERT_ALWAYS is that
// APD_ASSERT_ALWAYS_LIBRARY takes the file and line as parameters, so that
// the library package can pass them down from the original calling location.
#define APD_ASSERT_ALWAYS_LIBRARY(error_stream, file, line, test_expression, ...) \
    do \
      { \
         if ( !(test_expression) ) \
           { \
              fprintf(error_stream, "%s:%d: assertion failed: " \
                                    #test_expression, file, line); \
              fprintf(error_stream, "\n" __VA_ARGS__); \
              fprintf(error_stream, "\n"); \
              fprintf(error_stream, "assertion caught at %s:%d\n", __FILE__, __LINE__); \
              _APD_ABORT(error_stream, EXIT_FAILURE); \
           } \
      } while (0) /* swallow semicolon */

// this one is defined as a macro to make all the type casts transparent
#ifdef APD_ASSERT_COMPILE
#    define APD_ASSERT_BOUNDS(importance, expression, low_bound, high_bound, ...) \
          APD_CHECK_REGULARLY \
           do \
             { \
               APD_ASSERT(importance,  ((expression) >= (low_bound)), #expression ": %g is less than low bound of %g\n%s\n", \
                             (double)(expression), (double)(low_bound), "\n" __VA_ARGS__); \
               APD_ASSERT(importance,  ((expression) <= (high_bound)), #expression ": %g is more than high bound of %g\n%s\n", \
                             (double)(expression), (double)(high_bound), "\n" __VA_ARGS__); \
             } while (0) /* swallow semicolon */
#else
#    define APD_ASSERT_BOUNDS(importance, expression, low_bound, high_bound, ...)
#endif

#ifdef APD_ASSERT_COMPILE
#    define APD_ASSERT_BOUNDS_ALWAYS(expression, low_bound, high_bound, ...) \
          APD_CHECK_REGULARLY \
          APD_ASSERT_BOUNDS(0, expression, low_bound, high_bound, __VA_ARGS__)
#else
#    define APD_ASSERT_BOUNDS_ALWAYS(expression, low_bound, high_bound, ...) \
           do \
             { \
               APD_ASSERT_ALWAYS( ((expression) >= (low_bound)), #expression ": %g is less than low bound of %g\n%s\n", \
                             (double)(expression), (double)(low_bound), "\n" __VA_ARGS__); \
               APD_ASSERT_ALWAYS( ((expression) <= (high_bound)), #expression ": %g is more than high bound of %g\n%s\n", \
                             (double)(expression), (double)(high_bound), "\n" __VA_ARGS__); \
             } while (0) /* swallow semicolon */
#endif

// _APD_DIE_GUTS() is defined as a macro so that APD_DIE can be used
// without linking the apd library
#define _APD_DIE_GUTS(error_stream, file, line, function, ...) \
    fprintf(error_stream, APD_FILE_LINE_FORMAT, file, line); \
    fprintf(error_stream, APD_FUNCTION_NAME_FORMAT, function); \
    fprintf(error_stream, "" __VA_ARGS__); \
    _APD_ABORT(error_stream, EXIT_FAILURE)

// DIE macro wrapper: add file, line, and function name to apd_die message
#ifdef APD_EITHER_COMPILE
#    define APD_DIE(...)  \
          do { /* wrap in do-while in case this is used after an if() without enclosing {} */ \
              APD_CHECK_REGULARLY \
              _APD_DIE_GUTS(apd_object->error_stream, __FILE__, __LINE__, APD_FUNCTION_NAME, __VA_ARGS__); \
             } while(0) /* swallow semicolon */
#else
#    define APD_DIE(...) \
          do { \
              _APD_DIE_GUTS(APD_DEFAULT_ERROR_STREAM, __FILE__, __LINE__, APD_FUNCTION_NAME, __VA_ARGS__); \
             } while (0) /* swallow semicolon */
#endif

// ======================================
// ======= Function prototypes (external functions) ===
// ======================================
#ifdef NOTDEF
// apd_initialize() template - cut and paste me
apd_object = apd_initialize(
    /* int report_threshold */
    /* int assert_threshold */
    /* int max_function_depth */
    /* bool report_threshold_decreases_with_stack */
    /* FILE *error_stream */
    );
#endif

struct apd_object_struct *
apd_initialize(int report_threshold,
               int assert_threshold,
               uint max_function_depth,
               bool report_threshold_decreases_with_stack,
               FILE *error_stream);
void apd_end(struct apd_object_struct **apd_object_internal);
void apd_set(struct apd_object_struct *apd_object_internal, int set_type, int value);
void apd_set_file(struct apd_object_struct *apd_object_internal, int set_type, FILE *value);
#define APD_SET_REPORT_THRESHOLD  10
#define APD_SET_ASSERT_THRESHOLD  11
#define APD_SET_MAX_FUNCTION_DEPTH  12
#define APD_SET_ERROR_STREAM  13
#define APD_SET_INTERNAL_DEBUG  14
#define APD_SET_CHECK_INTERNAL  15
#define APD_SET_ACTION_EXTRA_INFO_ON  16
#define APD_SET_REPORT_THRESHOLD_DECREASES_WITH_STACK  17
#define APD_SET_VALGRIND_STACK_TRACE  18

// ======================================
// ======= Function prototypes (internal functions) ===
// ======================================
inline void _apd_report(struct apd_object_struct *apd_object_internal, char *file, int line, char *fmt, ...);
void _apd_pop_call_description(struct apd_object_struct *apd_object_internal, char *file, int line,
       const char *function_name);
void _apd_push_call_description(struct apd_object_struct *apd_object_internal, int importance, char *file,
      int line, const char *new_function_name, char *fmt, ...);
void _apd_report_function_stack(struct apd_object_struct *apd_object_internal, int threshold);
#endif
