#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>

// Redefine _APD_ABORT for testing:
#define _APD_TEST_MODE

#include <apd.h>

// ==========================================
// Testing system expanation:
// The testing system runs code under different conditions, and
// checks the output to make sure that
// 1. reports print when they should
// 2. reports do not print when they should not (or are not compiled in)
// 3. assertions cause an abort when they should
// 4. assertions do not cause an abort when they should not (or are not compiled in)
//
// Several conditions, such as "should print" or "did abort" are defined
// below.  Each condition is represented as a character.  The output
// is filtered so that only these characters remain, and then perl is
// used to make sure the conditions above hold.  See Makefile.am
// for the compile-time conditions tested.
//
// The function create_perl_test() creates a sh shell script that calls
// perl to do the dirty work of filtering and regular expressions.
// ==========================================
#define PERL_ESCAPE_CHARACTER "\\"

#define SHOULD_NOT_PRINT_BASE "&"
#define SHOULD_NOT_PRINT SHOULD_NOT_PRINT_BASE "<should not print>"
#define SHOULD_NOT_PRINT_PERL PERL_ESCAPE_CHARACTER SHOULD_NOT_PRINT_BASE

#define SHOULD_PRINT_BASE "^"
#define SHOULD_PRINT SHOULD_PRINT_BASE "<should print>"
#define SHOULD_PRINT_PERL PERL_ESCAPE_CHARACTER SHOULD_PRINT_BASE

#define DID_PRINT_BASE "%"
#define DID_PRINT DID_PRINT_BASE "<did print>"
#define DID_PRINT_PERL PERL_ESCAPE_CHARACTER DID_PRINT_BASE

#define SHOULD_NOT_ABORT_BASE "@"
#define SHOULD_NOT_ABORT SHOULD_NOT_ABORT_BASE "<should not abort>"
#define SHOULD_NOT_ABORT_PERL PERL_ESCAPE_CHARACTER SHOULD_NOT_ABORT_BASE

#define SHOULD_ABORT_BASE "#"
#define SHOULD_ABORT SHOULD_ABORT_BASE "<should abort>"
#define SHOULD_ABORT_PERL PERL_ESCAPE_CHARACTER SHOULD_ABORT_BASE

#define DID_ABORT_BASE "?"
#define DID_ABORT DID_ABORT_BASE "<did abort>"
#define DID_ABORT_PERL PERL_ESCAPE_CHARACTER DID_ABORT_BASE

#define ALL_PERL_INDICATORS \
    SHOULD_NOT_PRINT_PERL SHOULD_PRINT_PERL DID_PRINT_PERL \
    SHOULD_NOT_ABORT_PERL SHOULD_ABORT_PERL DID_ABORT_PERL

// Number of function level depths
#define TEST_LEVELS  25

// ========================================
// Create a Perl script to make sure the output of this test is OK
// ========================================
void create_perl_test(void)
{
FILE *test_file;

test_file = fopen("test.sh", "w");
APD_ASSERT_ALWAYS(test_file != NULL);
fprintf(test_file, "#!/bin/sh\n\n");
fprintf(test_file, "#\n# This file is generated by test_apd.c\n");

// Filter everything except the SHOULD/DID indicators
fprintf(test_file, "cat test.out | perl -p -e 's/[^" ALL_PERL_INDICATORS "]//sg;' > test.out2 \n");
// Search for instances of no DID_PRINT after a SHOULD_PRINT
fprintf(test_file, "cat test.out2 | perl -p -e '/" SHOULD_PRINT_PERL "[^" DID_PRINT_PERL "]/g or $_=\"\";' > test_result.1 \n");
fprintf(test_file, "if [ ! -z \"$(cat test_result.1)\" ]; then echo \"found non-DID_PRINT after SHOULD_PRINT\"; exit 1; fi \n");
// Search for instances of a DID_PRINT directly after a SHOULD_NOT_PRINT
fprintf(test_file, "cat test.out2 | perl -p -e '/" SHOULD_NOT_PRINT_PERL DID_PRINT_PERL "/g or $_=\"\";' > test_result.2 \n");
fprintf(test_file, "if [ ! -z \"$(cat test_result.2)\" ]; then echo \"found DID_PRINT after SHOULD_NOT_PRINT\"; exit 1; fi \n");

// Search for instances of no DID_ABORT after a SHOULD_ABORT
fprintf(test_file, "cat test.out2 | perl -p -e '/" SHOULD_ABORT_PERL "[^" DID_ABORT_PERL "]/g or $_=\"\";' > test_result.3 \n");
fprintf(test_file, "if [ ! -z \"$(cat test_result.3)\" ]; then echo \"found non-DID_ABORT after SHOULD_ABORT\"; exit 1; fi \n");
// Search for instances of a DID_ABORT directly after a SHOULD_NOT_ABORT
fprintf(test_file, "cat test.out2 | perl -p -e '/" SHOULD_NOT_ABORT_PERL DID_ABORT_PERL "/g or $_=\"\";' > test_result.4 \n");
fprintf(test_file, "if [ ! -z \"$(cat test_result.4)\" ]; then echo \"found DID_ABORT after SHOULD_NOT_ABORT\"; exit 1; fi \n");

fclose(test_file);
} // end create_perl_test()

// ========================================
// ========================================
void call_recurse(int level)
{
APD_PUSH_FUNCTION(5, "level=%d", level);
if (level > 1) call_recurse(level-1);
APD_POP_FUNCTION();
} // end call_recurse()

// ========================================
// ========================================
#undef APD_ADJUST_REPORT_THRESHOLD
#define APD_ADJUST_REPORT_THRESHOLD 1
void call_adjusted(int threshold)
{
APD_PUSH_FUNCTION(0, "Starting call_adjusted with threshold=%d", threshold);

#ifdef APD_REPORT_COMPILE
    printf(SHOULD_PRINT);
#else
    printf(SHOULD_NOT_PRINT);
#endif
APD_REPORT(threshold, DID_PRINT);

APD_POP_FUNCTION();
} // end call_adjusted()
#undef APD_ADJUST_REPORT_THRESHOLD
#define APD_ADJUST_REPORT_THRESHOLD 0

// ========================================
// ========================================
// This main() is only for running the tests in this file
// when compiled with "make test"
int main(void)
{
apd_object = apd_initialize(
    3,  /* int report_threshold */
    3,  /* int assert_threshold */
    TEST_LEVELS,  /* int max_function_depth */
    0,  /* int report_threshold_decreases_with_stack */
    stdout  /* FILE *error_stream) */
    );

// To debug the apd library, do this:
apd_set(apd_object, APD_SET_INTERNAL_DEBUG, 1);
// To perform internal consistency checks on apd data structures continuously:
apd_set(apd_object, APD_SET_CHECK_INTERNAL, 1);

APD_PUSH_FUNCTION(3);

// ========================
// Test function tracking with a recursive function almost
// up to the maximum TEST_LEVELS deep.
printf("Testing function tracking\n");
call_recurse(TEST_LEVELS-1);

// ========================
// Report Macros
// ========================

// ========================
printf("Testing ALD_APD_REPORT macro\n");  fflush(stdout);
#ifdef APD_REPORT_COMPILE
    printf(SHOULD_PRINT);
#else
    printf(SHOULD_NOT_PRINT);
#endif
APD_REPORT(3, DID_PRINT "importance = 3");
printf(SHOULD_NOT_PRINT);
APD_REPORT(4, DID_PRINT "importance = 4");

printf("Testing APD_REPORT macro in a function adjusted with APD_ADJUST_REPORT_THRESHOLD\n");
call_adjusted(3);
call_adjusted(4);

// ========================
printf("Testing APD_REPORT_ALWAYS macro\n");
printf(SHOULD_PRINT);
APD_REPORT_ALWAYS(DID_PRINT);

// ========================
printf("Testing APD_REPORT_OTHERWISE macro\n");
printf(SHOULD_NOT_PRINT);
APD_REPORT_OTHERWISE(3, DID_PRINT "importance = 3");
#ifdef APD_REPORT_COMPILE
    printf(SHOULD_PRINT);
#else
    printf(SHOULD_NOT_PRINT);
#endif
APD_REPORT_OTHERWISE(4, DID_PRINT "importance = 4");

// ========================
// Assert Macros
// ========================

// ========================
printf("Testing APD_ASSERT\n");
printf(SHOULD_NOT_ABORT);
APD_ASSERT(3, 1>0);
printf(SHOULD_NOT_ABORT);
APD_ASSERT(4, 0>1);
#ifdef APD_ASSERT_COMPILE
    printf(SHOULD_ABORT);
#else
    printf(SHOULD_NOT_ABORT);
#endif
APD_ASSERT(3, 0>1);

// ========================
printf("Testing APD_ASSERT_ALWAYS\n");
printf(SHOULD_NOT_ABORT);
APD_ASSERT_ALWAYS(1>0);
printf(SHOULD_ABORT);
APD_ASSERT_ALWAYS(0>1);

// ========================
printf("Testing APD_ASSERT_BOUNDS\n");
printf(SHOULD_NOT_ABORT);
APD_ASSERT_BOUNDS(3, 30, 30, 50, "Out of bounds, loser");
APD_ASSERT_BOUNDS(3, 50, 30, 50, "Out of bounds, loser");
APD_ASSERT_BOUNDS(4, 29, 30, 50, "Out of bounds, loser");
APD_ASSERT_BOUNDS(4, 51, 30, 50, "Out of bounds, loser");
#ifdef APD_ASSERT_COMPILE
    printf(SHOULD_ABORT);
#else
    printf(SHOULD_NOT_ABORT);
#endif
APD_ASSERT_BOUNDS(3, 29, 30, 50, "Out of bounds, loser");
#ifdef APD_ASSERT_COMPILE
    printf(SHOULD_ABORT);
#else
    printf(SHOULD_NOT_ABORT);
#endif
APD_ASSERT_BOUNDS(3, 29, 30, 51, "Out of bounds, loser");

// ========================
printf("Testing APD_ASSERT_BOUNDS_ALWAYS\n");
printf(SHOULD_NOT_ABORT);
APD_ASSERT_BOUNDS_ALWAYS(30, 30, 50, "Out of bounds, loser");
APD_ASSERT_BOUNDS_ALWAYS(50, 30, 50, "Out of bounds, loser");
printf(SHOULD_ABORT);
APD_ASSERT_BOUNDS_ALWAYS(29, 30, 50, "Out of bounds, loser");
printf(SHOULD_ABORT);
APD_ASSERT_BOUNDS_ALWAYS(51, 30, 50, "Out of bounds, loser");

// ========================
// Misc Macros
// ========================

// ========================
printf("Testing APD_REPORT_ACTION macro\n");
#ifdef APD_REPORT_COMPILE
    printf(SHOULD_PRINT);
#else
    printf(SHOULD_NOT_PRINT);
#endif
APD_REPORT_ACTION(3, printf(DID_PRINT); );
printf(SHOULD_NOT_PRINT);
APD_REPORT_ACTION(4, printf(DID_PRINT); );

// ========================
printf("Testing APD_ASSERT_ACTION macro\n");
#ifdef APD_ASSERT_COMPILE
    printf(SHOULD_PRINT);
#else
    printf(SHOULD_NOT_PRINT);
#endif
APD_ASSERT_ACTION(3, (1 == 0), printf(DID_PRINT); );
printf(SHOULD_NOT_PRINT);
APD_ASSERT_ACTION(3, (1 == 1), printf(DID_PRINT); );
printf(SHOULD_NOT_PRINT);
APD_ASSERT_ACTION(4, (1 == 0), printf(DID_PRINT); );

// ========================
printf("Testing APD_ACTION macro\n");
#ifdef APD_ASSERT_COMPILE
    printf(SHOULD_PRINT);
#else
    printf(SHOULD_NOT_PRINT);
#endif
APD_ACTION(3, printf(DID_PRINT); );
printf(SHOULD_NOT_PRINT);
APD_ACTION(4, printf(DID_PRINT); );

// ========================
printf("Testing APD_DIE macro\n");
printf(SHOULD_ABORT);
APD_DIE("Loser.");

create_perl_test();
APD_POP_FUNCTION();
apd_end(&apd_object);
return(EXIT_SUCCESS);
} // end main()

